package org.lazywizard.omnifac;

import java.util.ArrayList;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.campaign.*;
import com.fs.starfarer.api.campaign.econ.SubmarketAPI;
import com.fs.starfarer.api.combat.ShipHullSpecAPI;
import com.fs.starfarer.api.combat.WeaponAPI;
import com.fs.starfarer.api.fleet.FleetMemberAPI;
import com.fs.starfarer.api.fleet.FleetMemberType;
import com.fs.starfarer.api.impl.campaign.ids.Commodities;
import com.fs.starfarer.api.impl.campaign.ids.Submarkets;
import com.fs.starfarer.api.impl.campaign.submarkets.StoragePlugin;
import com.fs.starfarer.api.loading.FighterWingSpecAPI;
import com.fs.starfarer.api.loading.WeaponSlotAPI;
import com.fs.starfarer.api.loading.WeaponSpecAPI;

import org.apache.log4j.Logger;
import org.lazywizard.lazylib.CollectionUtils;
import org.lazywizard.lazylib.campaign.MessageUtils;
import org.lazywizard.omnifac.secondary.Constants;

public abstract class Shipyard extends StoragePlugin
{
	// Strings
    protected static final String SHIP_NAME = "ship", WEAPON_NAME = "weapon", WING_NAME = "LPC";
	protected static final String NO_REPLICATE = "Unable to Replicate.";
	protected static final String HAS_BLUEPRINT = "Blueprint already known";
	protected static final String GAMMA_CORE_NEEDED = "Requires Gamma Class AI Core.";
	protected static final String GAMMA_CORES_NEEDED = "Requires 2 Gamma Class AI Cores.";
	protected static final String BETA_CORE_NEEDED = "Requires Beta Class AI Core.";
	protected static final String ALPHA_CORE_NEEDED = "Requires Alpha Class AI Core.";
	protected static final String ALPHA_CORES_NEEDED = "Requires 5 Alpha Class AI Cores.";
	// Privates
	private static Logger log = Global.getLogger(Shipyard.class);
    private final Map<String, Data> shipData = new HashMap<>();
	private final Map<String, Data> wepData = new HashMap<>();
	private final Map<String, Data> wingData = new HashMap<>();
    private SectorEntityToken station;
    private boolean displayCosts = false;
	private boolean warnedRequirements = true, refit = false;
    private long lastHeartbeat;
    private int numHeartbeats = 0;

    @Override
    public void init(SubmarketAPI submarket)
    {
        super.init(submarket);
        super.setPlayerPaidToUnlock(true);
//	    CampaignEngine.getInstance().getCampaignUI().getCore().setListener(new AutofactoryInteractionDialog());

        // Will be properly set by initOmnifactory(), but here as a fallback
        this.station = submarket.getMarket().getPrimaryEntity();

        // Synchronize factory heartbeat to the start of the next day
        final CampaignClockAPI clock = Global.getSector().getClock();
        lastHeartbeat = new GregorianCalendar(clock.getCycle(),
                clock.getMonth() - 1, clock.getDay()).getTimeInMillis();
    }

    public List<String> getKnownShips()
    {
        return new ArrayList<>(getShipData().keySet());
    }

    public List<String> getKnownWings()
    {
        return new ArrayList<>(getWingData().keySet());
    }

    public List<String> getKnownWeapons()
    {
        return new ArrayList<>(getWepData().keySet());
    }

    public Data getShipBlueprint(String hullOrWingId)
    {
        return getShipData().get(hullOrWingId);
    }

    public Data getWeaponBlueprint(String weaponId)
    {
        return getWepData().get(weaponId);
    }

    public SubmarketAPI getStorageMarket()
    {
        if (!market.hasSubmarket(Submarkets.SUBMARKET_STORAGE))
        {
            market.addSubmarket(Submarkets.SUBMARKET_STORAGE);
            ((StoragePlugin) market.getSubmarket(Submarkets.SUBMARKET_STORAGE)
                    .getPlugin()).setPlayerPaidToUnlock(true);
            market.addSubmarket(Constants.SUBMARKET_ID);
        }

        return market.getSubmarket(Submarkets.SUBMARKET_STORAGE);
    }

    public CargoAPI getFactoryCargo()
    {
        return getCargo();
    }

    public CargoAPI getStorageCargo()
    {
        return getStorageMarket().getCargo();
    }

    public SectorEntityToken getStation()
    {
        return station;
    }

	protected void setStation(SectorEntityToken station)
	{
		this.station = station;
	}

    public String getLocationString()
    {
        OrbitAPI orbit = station.getOrbit();
        LocationAPI loc = station.getContainingLocation();
        if (orbit == null || orbit.getFocus() == null)
        {
            return "orbiting nothing in " + (loc == null ? " nowhere" : loc.getName());
        }

        return "orbiting " + orbit.getFocus().getName() + " in " + loc.getName();
    }

    @Override
    public String toString()
    {
        return station.getName() + " " + getLocationString()
                + " (" + getShipData().size() + " ships, " + getWepData().size()
                + " weapons known)";
    }

    protected boolean checkTimerAnalyze(Data tmp)
    {
	    return numHeartbeats - tmp.getDaysToAnalyze() >= tmp.getLastUpdate();
    }

	protected boolean checkTimerCreate(Data tmp)
	{
		return numHeartbeats - tmp.getDaysToCreate() >= tmp.getLastUpdate();
	}

    protected abstract void processBlueprint(Data tmp, List<String> analyzed, List<String> added, List<String> limit, String name);
	protected abstract void addBluePrint(List<String> newItems, String name);
	protected abstract boolean blockBluePrint(List<String> blocked, String name);
	protected abstract void orderedMessage(List<String> ordered, String name);
	protected abstract void producedMessage(List<String> added, String name);
	protected abstract void finishMessage(List<String> analyzed, String name);

    private void heartbeat()
    {
        numHeartbeats++;

        List<String> addedShips = new ArrayList<>(), addedWeps = new ArrayList<>(), addedWings = new ArrayList<>(),
                analyzedShips = new ArrayList<>(), analyzedWeps = new ArrayList<>(), analyzedWings = new ArrayList<>(),
                hitLimit = new ArrayList<>();

        for (Data tmp : getShipData().values())
            processBlueprint(tmp, analyzedShips, addedShips, hitLimit, SHIP_NAME);
        for (Data tmp : getWepData().values())
            processBlueprint(tmp, analyzedWeps, addedWeps, hitLimit, WEAPON_NAME);
        for (Data tmp : getWingData().values())
            processBlueprint(tmp, analyzedWings, addedWings, hitLimit, WING_NAME);

        if (OmniFacSettings.shouldShowAddedCargo())
        {
            if (!addedShips.isEmpty())
                producedMessage(addedShips, SHIP_NAME);
            if (!addedWeps.isEmpty())
                producedMessage(addedWeps, WEAPON_NAME);
            if (!addedWings.isEmpty())
                producedMessage(addedWings, WING_NAME);
        }

        if (OmniFacSettings.shouldShowLimitReached() && !hitLimit.isEmpty())
        {
            Collections.sort(hitLimit);
            MessageUtils.showMessage("The " + station.getName()
                    + " has reached its limit for the following goods:",
                    CollectionUtils.implode(hitLimit) + ".", true);
        }

        if (OmniFacSettings.shouldShowAnalysisComplete())
        {
            if (!analyzedShips.isEmpty())
                finishMessage(analyzedShips, SHIP_NAME);
            if (!analyzedWeps.isEmpty())
                finishMessage(analyzedWeps, WEAPON_NAME);
            if (!analyzedWings.isEmpty())
                finishMessage(analyzedWings, WING_NAME);
        }
    }

    protected void Warning(Data tmp, List<String> addedItems, List<String> hitLimit) throws NullPointerException
    {
        if (utilCreate(tmp, true))
        {
            if (OmniFacSettings.shouldShowAddedCargo())
                addedItems.add(tmp.getDisplayName() + " (" + tmp.getTotal()
                        + "/" + (tmp.getTotal() + tmp.getProduction()) + ")");
        }
        else if (OmniFacSettings.shouldShowLimitReached() && !tmp.hasWarnedLimit())
        {
            hitLimit.add(tmp.getDisplayName());
            tmp.setWarnedLimit(true);
        }
    }

    protected boolean utilCreate(Data tmp, boolean isReal)
    {
    	try {
		    return tmp.create(isReal);
	    } catch (NullPointerException e) {
    		// TODO: Integrate log4j
			return false;
	    }
    }

    @Override
    public void advance(float amount)
    {
        CampaignClockAPI clock = Global.getSector().getClock();
        refit = false;

        if (clock.getElapsedDaysSince(lastHeartbeat) >= 1f)
        {
            lastHeartbeat = clock.getTimestamp();
            heartbeat();
        }
    }

    @Override
    public OnClickAction getOnClickAction(CoreUIAPI ui) {
	    return OnClickAction.OPEN_SUBMARKET;
    }

    @Override
    public String getName()
    {
        return Constants.SUBMARKET_NAME;
    }

    @Override
    public String getBuyVerb()
    {
        return "Order";
    }

    @Override
    public String getSellVerb()
    {
        return "Analyse";
    }

    @Override
    public boolean isIllegalOnSubmarket(String commodityId, TransferAction action)
    {
        return true;
    }

    @Override
    public boolean isIllegalOnSubmarket(CargoStackAPI stack, TransferAction action)
    {
        return action == TransferAction.PLAYER_SELL;
    }

    @Override
    public boolean isIllegalOnSubmarket(FleetMemberAPI ship, TransferAction action)
    {
	    return action == TransferAction.PLAYER_SELL;
    }

    @Override
    public float getTariff()
    {
        return OmniFacSettings.getOmnifactoryTariff();
    }

    @Override
    public boolean isFreeTransfer()
    {
        return false;
    }

    @Override
    public boolean isParticipatesInEconomy()
    {
        return false;
    }

    @Override
    public void updateCargoPrePlayerInteraction()
    {
        // TODO: Modify demand
    }

	public Map<String, Data> getShipData()
	{
		return shipData;
	}

	public Map<String, Data> getWepData()
	{
		return wepData;
	}

	public Map<String, Data> getWingData()
	{
		return wingData;
	}

	public void setRefit(boolean refit)
	{
		this.refit = refit;
	}

	public boolean isRefit()
	{
		return refit;
	}

	protected boolean isDisplayCosts()
	{
		return displayCosts;
	}

	public abstract class Data
    {
        String id, displayName;
        float size;
        int ordered, lastUpdate, stackSize, goal;
        boolean warnedLimit = false, isAnalyzed = false;
        HashMap<String, Integer> cost;

        public abstract int getDaysToAnalyze();

        protected abstract int getBaseDays();

        public abstract int getDaysToCreate();

        public int getLastUpdate()
        {
            return lastUpdate;
        }

        public String getDisplayName()
        {
            return displayName;
        }

        public String getId()
        {
            return id;
        }

        public int getTotal()
        {
	        return goal - ordered;
        }

        public abstract int getLimit();

        public boolean inProduction() { return ordered > 0; }

        public int getProduction() {
        	if (ordered <= 0)
        		goal = 0;
        	return ordered;
        }

        public void startProduction(int requested)
        {
        	boolean canAfford = true;
        	for (Map.Entry<String, Integer> commodity : cost.entrySet())
		        canAfford = canAfford && getStorageCargo().getQuantity(CargoAPI.CargoItemType.RESOURCES, commodity.getKey()) >= commodity.getValue() * requested;
        	if (canAfford)
	        {
		        removeMaterials(requested);
		        goal = ordered = requested;
		        lastUpdate = numHeartbeats;
	        } else
	        {
		        Global.getSector().getCampaignUI().addMessage("Insufficient resources, order cancelled.");
	        }
        }

        public HashMap<String, Integer> getCost()
        {
            return cost;
        }

        public boolean hasWarnedLimit()
        {
            return warnedLimit;
        }

        public void setWarnedLimit(boolean hasWarned)
        {
            warnedLimit = hasWarned;
        }

        public boolean isAnalyzed()
        {
            return isAnalyzed;
        }

        public void setAnalyzed(boolean isAnalyzed)
        {
            this.isAnalyzed = isAnalyzed;
            lastUpdate = numHeartbeats;
        }

        public void removeMaterials(int mult)
        {
            for (Map.Entry<String, Integer> commodity : cost.entrySet())
            {
            	if (commodity.getValue() * mult > 0)
	                Global.getSector().getCampaignUI().addMessage("Consumed " + commodity.getValue() * OmniFacSettings.getCostMult() * mult + " " + Global.getSector().getEconomy().getCommoditySpec(commodity.getKey()).getName());
	            getStorageCargo().removeCommodity(commodity.getKey(), commodity.getValue() * OmniFacSettings.getCostMult() * mult);
            }
        }

        public abstract boolean create(boolean really) throws NullPointerException;
    }

	protected class ShipData extends Data
    {
        FleetMemberType type;
        int fp, supplies;

        ShipData(FleetMemberAPI ship)
        {
	        ordered = 0;
	        goal = 0;
            cost = new HashMap<>();
            ShipHullSpecAPI spec = ship.getHullSpec();
            int weapons = 0;
            for (WeaponSlotAPI slot: spec.getAllWeaponSlotsCopy())
                switch (slot.getSlotSize()) {
                    case SMALL:
                        weapons += 10;
                        break;
                    case MEDIUM:
                        weapons += 20;
                        break;
                    case LARGE:
                        weapons += 40;
                        break;
                    default:
                        break;
                }
	        String rawid = ship.getHullId();
            id = rawid + (rawid.contains("default_D") ? "" : "_Hull");
//            id = ship.getHullId() + "_Hull";
            displayName = OmniFacUtil.parseHullName(ship);
            type = ship.getType();
            fp = ship.getFleetPointCost();
            size = ship.getHullSpec().getHullSize().ordinal();
	        supplies = (int) ship.getBaseDeploymentCostSupplies();
            lastUpdate = numHeartbeats;
            cost.put(Commodities.METALS, (int)(spec.getArmorRating() + spec.getHitpoints() / 10));
            cost.put(Commodities.HEAVY_MACHINERY, (int)(spec.getFighterBays() * 50 + weapons));
            cost.put(Commodities.RARE_METALS, (int)(spec.getOrdnancePoints(null) + spec.getArmorRating() / 50 + spec.getFluxCapacity() / 500 + spec.getFluxDissipation() / 50));
            cost.put(Commodities.ORGANICS, (int)(spec.getMaxCrew() / 10));
            cost.put(Commodities.VOLATILES, (int)(spec.getShieldSpec().getUpkeepCost() + spec.getShieldSpec().getPhaseUpkeep()
                    + (spec.getShieldSpec().getFluxPerDamageAbsorbed() != 0 ? (1 / spec.getShieldSpec().getFluxPerDamageAbsorbed()) * 100 : 0) + spec.getShieldSpec().getPhaseCost()));
        }

        @Override
        public int getDaysToAnalyze()
        {
            return (int) Math.max(1f,
		            getBaseDays() * OmniFacSettings.getShipAnalysisTimeMod());
        }

        @Override
	    protected int getBaseDays()
        {
	        return fp + supplies;
        }

        @Override
        public int getDaysToCreate()
        {
            return (int) Math.max(1f,
		            getBaseDays() * OmniFacSettings.getShipProductionTimeMod());
        }

//        @Override
//        public int getTotal()
//        {
//            int total = 0;
//
//            for (FleetMemberAPI tmp : getStorageCargo().getMothballedShips().getMembersListCopy())
//                if (id.equals(parseHullName(tmp)))
//                    total++;
//
//            return total;
//        }

        @Override
        public int getLimit()
        {
            return 1;
        }

        @Override
        public boolean create(boolean really) throws NullPointerException
        {
            lastUpdate = numHeartbeats;

            if (getTotal() >= getLimit())
            {
                return false;
            }

            warnedLimit = false;
            if (really)
            {
                getStorageCargo().addMothballedShip(type, id,null);
	            ordered--;
            } else
	            getFactoryCargo().addMothballedShip(type, id,null);
            return true;
        }
    }

	protected class WeaponData extends Data
    {
        int tier;
        WeaponData(CargoStackAPI stack)
        {
            ordered = 0;
            cost = new HashMap<>();
            WeaponSpecAPI spec = stack.getWeaponSpecIfWeapon();
            id = (String) stack.getData();
            displayName = stack.getDisplayName();
            tier = spec.getTier() + 1;
            size = stack.getCargoSpacePerUnit();
            //2 40
            //4 20
            //8 10
            stackSize = (int) (80 / size);
            lastUpdate = numHeartbeats;
            cost.put(Commodities.METALS, (int)(size * 2));
            cost.put(Commodities.FUEL, (int)(spec.getType() == WeaponAPI.WeaponType.MISSILE ? size * spec.getTier() : 0));
            cost.put(Commodities.RARE_METALS, (int)(spec.getType() == WeaponAPI.WeaponType.BALLISTIC ? size * spec.getTier() : 0));
            cost.put(Commodities.VOLATILES, (int)(spec.getType() == WeaponAPI.WeaponType.ENERGY ? size * 2 * spec.getTier() : 0));
        }

        @Override
        public int getDaysToAnalyze()
        {
            return (int) Math.max(1f,
		            getBaseDays() * OmniFacSettings.getWeaponAnalysisTimeMod());
        }

	    @Override
	    protected int getBaseDays()
	    {
		    return (int)(size * tier);
	    }

        @Override
        public int getDaysToCreate()
        {
            return (int) Math.max(getBaseDays() * OmniFacSettings.getWeaponProductionTimeMod(), 1f);
        }

//        @Override
//        public int getTotal()
//        {
//            return getStorageCargo().getNumWeapons(id);
//        }

        @Override
        public int getLimit()
        {
            return (int) (stackSize * OmniFacSettings.getMaxStacksPerWeapon());
        }

        @Override
        public boolean create(boolean really) throws NullPointerException
        {
            lastUpdate = numHeartbeats;

            if (getTotal() >= getLimit())
            {
                return false;
            }

            warnedLimit = false;
            if (really)
            {
                getStorageCargo().addWeapons(id, 1);
	            ordered--;
            }
            else
                getFactoryCargo().addWeapons(id, getLimit() - getFactoryCargo().getNumWeapons(id));
            return true;
        }
    }

    protected class WingData extends Data
    {
        int tier;
        WingData(CargoStackAPI stack)
        {
            ordered = 0;
            cost = new HashMap<>();
            FighterWingSpecAPI spec = stack.getFighterWingSpecIfWing();
            id = (String) stack.getData();
            displayName = stack.getDisplayName();
            tier = spec.getTier() + 1;
            size = stack.getCargoSpacePerUnit() * 10;
            //1 10
            stackSize = (int) (50 / size);
            lastUpdate = numHeartbeats;
	        cost.put(Commodities.RARE_METALS, (int)(size * spec.getTier()));
        }

        @Override
        public int getDaysToAnalyze()
        {
            return (int) Math.max(1f,
		            getBaseDays() * OmniFacSettings.getWingAnalysisTimeMod());
        }

	    @Override
	    protected int getBaseDays()
	    {
		    return (int)(size * tier);
	    }

        @Override
        public int getDaysToCreate()
        {
            return (int) Math.max(getBaseDays() * OmniFacSettings.getWingProductionTimeMod(), 1f);
        }

//        @Override
//        public int getTotal()
//        {
//            return getStorageCargo().getNumFighters(id) - goal;
//        }

        @Override
        public int getLimit()
        {
            return (int) (stackSize * OmniFacSettings.getMaxStacksPerWing());
        }

        @Override
        public boolean create(boolean really)
        {
            lastUpdate = numHeartbeats;

            if (getTotal() >= getLimit())
            {
                return false;
            }

            warnedLimit = false;
            if (really)
            {
                getStorageCargo().addFighters(id, 1);
	            ordered--;
            } else
                getFactoryCargo().addFighters(id, getLimit() - getFactoryCargo().getNumFighters(id));
            return true;
        }
    }
}
